#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <dos.h>
#include <malloc.h>

#include <a/types.h>
#include <a/vexe.h>

////////////////////////

void	Warning(char *s, ...);
void	Error(char *s, ...);

////////////////////////////////////////////////////////////////////////////

#define N		 4096	/* size of ring buffer */
#define F		   18	/* upper limit for match_length */
#define F2		   83	/* upper limit for match_length */
#define THRESHOLD	2   /* encode string into position and length
				if match_length is greater than this */

#define NIL			N	/* index for root of binary search trees */

unsigned long int
		textsize,	/* text size counter */
		codesize,	/* code size counter */
		printcount;	/* counter for reporting progress every 1K bytes */
unsigned char
		text_buf[N + F2 - 1];	/* ring buffer of size N,
			with extra F-1 bytes to facilitate string comparison */
int		match_position, match_length,  /* of longest match.  These are
			set by the InsertNode() procedure. */
		lson[N + 1], rson[N + 257], dad[N + 1];  /* left & right children &
			parents -- These constitute binary search trees. */

void InitTree(void)  /* initialize trees */
{
	int  i;

	/* For i = 0 to N - 1, rson[i] and lson[i] will be the right and
	   left children of node i.  These nodes need not be initialized.
	   Also, dad[i] is the parent of node i.  These are initialized to
	   NIL (= N), which stands for 'not used.'
	   For i = 0 to 255, rson[N + i + 1] is the root of the tree
	   for strings that begin with character i.  These are initialized
	   to NIL.  Note there are 256 trees. */

	for (i = N + 1; i <= N + 256; i++) rson[i] = NIL;
	for (i = 0; i < N; i++) dad[i] = NIL;
}

void InsertNode(int r)
	/* Inserts string of length F, text_buf[r..r+F-1], into one of the
	   trees (text_buf[r]'th tree) and returns the longest-match position
	   and length via the global variables match_position and match_length.
	   If match_length = F, then removes the old node in favor of the new
	   one, because the old one will be deleted sooner.
	   Note r plays double role, as tree node and position in buffer. */
{
	int  i, p, cmp;
	unsigned char  *key;

	cmp = 1;  key = &text_buf[r];  p = N + 1 + key[0];
	rson[r] = lson[r] = NIL;  match_length = 0;
	for ( ; ; ) {
		if (cmp >= 0) {
			if (rson[p] != NIL) p = rson[p];
			else {  rson[p] = r;  dad[r] = p;  return;  }
		} else {
			if (lson[p] != NIL) p = lson[p];
			else {  lson[p] = r;  dad[r] = p;  return;  }
		}
		for (i = 1; i < F; i++)
			if ((cmp = key[i] - text_buf[p + i]) != 0)  break;
		if (i > match_length) {
			match_position = p;
			if ((match_length = i) >= F)  break;
		}
	}
	dad[r] = dad[p];  lson[r] = lson[p];  rson[r] = rson[p];
	dad[lson[p]] = r;  dad[rson[p]] = r;
	if (rson[dad[p]] == p) rson[dad[p]] = r;
	else                   lson[dad[p]] = r;
	dad[p] = NIL;  /* remove p */
}

void InsertNode2(int r)
	/* Inserts string of length F, text_buf[r..r+F-1], into one of the
	   trees (text_buf[r]'th tree) and returns the longest-match position
	   and length via the global variables match_position and match_length.
	   If match_length = F, then removes the old node in favor of the new
	   one, because the old one will be deleted sooner.
	   Note r plays double role, as tree node and position in buffer. */
{
	int  i, p, cmp;
	unsigned char  *key;

	cmp = 1;  key = &text_buf[r];  p = N + 1 + key[0];
	rson[r] = lson[r] = NIL;  match_length = 0;
	for ( ; ; ) {
		if (cmp >= 0) {
			if (rson[p] != NIL) p = rson[p];
			else {  rson[p] = r;  dad[r] = p;  return;  }
		} else {
			if (lson[p] != NIL) p = lson[p];
			else {  lson[p] = r;  dad[r] = p;  return;  }
		}
		for (i = 1; i < F2; i++)
			if ((cmp = key[i] - text_buf[p + i]) != 0)  break;
		if (i > match_length) {
			match_position = p;
			if ((match_length = i) >= F2)  break;
		}
	}
	dad[r] = dad[p];  lson[r] = lson[p];  rson[r] = rson[p];
	dad[lson[p]] = r;  dad[rson[p]] = r;
	if (rson[dad[p]] == p) rson[dad[p]] = r;
	else                   lson[dad[p]] = r;
	dad[p] = NIL;  /* remove p */
}

void DeleteNode(int p)  /* deletes node p from tree */
{
	int  q;
	
	if (dad[p] == NIL) return;  /* not in tree */
	if (rson[p] == NIL) q = lson[p];
	else if (lson[p] == NIL) q = rson[p];
	else {
		q = lson[p];
		if (rson[q] != NIL) {
			do {  q = rson[q];  } while (rson[q] != NIL);
			rson[dad[q]] = lson[q];  dad[lson[q]] = dad[q];
			lson[q] = lson[p];  dad[lson[p]] = q;
		}
		rson[q] = rson[p];  dad[rson[p]] = q;
	}
	dad[q] = dad[p];
	if (rson[dad[p]] == p) rson[dad[p]] = q;  else lson[dad[p]] = q;
	dad[p] = NIL;
}

unsigned long Encode(char *inmem,char *outmem,long inbytes)
{
	int  i, c, len, r, s, last_match_length, code_buf_ptr;
	unsigned char  code_buf[17], mask;
	unsigned long codes=0;

	unsigned short index=0x10000-F;

	char *start = outmem;
	
	outmem += 4;

	codesize = printcount = 0;

	InitTree();
	code_buf[0] = 0;

	code_buf_ptr = mask = 1;
	s = 0;  r = N - F;
	for (i = s; i < r; i++) text_buf[i] = 0;
	for (len = 0; len < F; len++) {
		if (inbytes--<=0) break;
		text_buf[r + len] = *inmem++;
	}
	if ((textsize = len) == 0) {
		*(unsigned long *)start = 0;
		return 0;
	}
	for (i = 1; i <= F; i++) InsertNode(r - i);
	InsertNode(r);
	do {
		if (match_length > len) match_length = len;
		if (match_length <= THRESHOLD) {
			match_length = 1;
			code_buf[0] |= mask;
			code_buf[code_buf_ptr++] = text_buf[r];
		} else {
			match_position-=index;
			code_buf[code_buf_ptr++] = (unsigned char) match_position;
			code_buf[code_buf_ptr++] = (unsigned char)
				(((match_position >> 4) & 0xf0)
			  | (match_length - (THRESHOLD + 1)));
		}
		++codes;
		index+=match_length;
		if ((mask <<= 1) == 0) {
			for (i = 0; i < code_buf_ptr; i++) *outmem++ = code_buf[i];
			codesize += code_buf_ptr;
			code_buf[0] = 0;  code_buf_ptr = mask = 1;
		}
		last_match_length = match_length;
		for (i = 0; i < last_match_length; i++) {
			if (inbytes--<=0) break;
			DeleteNode(s);
			text_buf[s] = *inmem++;

			if (s < F - 1) text_buf[s + N] = c;
			s = (s + 1) & (N - 1);  r = (r + 1) & (N - 1);
			InsertNode(r);
		}
		textsize += i;
		while (i++ < last_match_length) {
			DeleteNode(s);
			s = (s + 1) & (N - 1);  r = (r + 1) & (N - 1);
			if (--len) InsertNode(r);
		}
	} while (len > 0);
	if (code_buf_ptr > 1) {
		for (i = 0; i < code_buf_ptr; i++) *outmem++ = code_buf[i];
		codesize += code_buf_ptr;
	}

//	printf("%ld -> %ld\n", textsize,codesize);

	*(unsigned long *)start = codes;

	return outmem - start;
}

/////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////

typedef	short	HuffInt;

static int HuffmanHistoPassCompare(const HuffInt *i1,const HuffInt *i2);
static void HuffmanBuildTree(ushort *e_table,uchar *bitlen_table,struct HuffmanData *hd_ptr,int maxc);
static void HistoAssignBitlen(int inode,int blen);

#define MAX_LEN (15)

static struct HuffmanData {
	ulong	histo[560];
	HuffInt	left[560];
	HuffInt	right[560];
	HuffInt	tmatrix[560];
	HuffInt	bitlcnt[MAX_LEN+1];
	char	bitlen[280];
	char	tdepth[560];

	int	histo_overflow;
} *hdp;

static int HuffmanHistoPassCompare(const HuffInt *i1,const HuffInt *i2) {
	HuffInt n1,n2;

	n1 = *i1;
	n2 = *i2;

	if (hdp->histo[n1]<hdp->histo[n2]) return 1;
	else if (hdp->histo[n1]>hdp->histo[n2]) return -1;
	else if (hdp->tdepth[n1]<hdp->tdepth[n2]) return 1;
	else return -1;
}

static void	HuffmanBuildTree(ushort *e_table,uchar *bitlen_table,struct HuffmanData *hd_ptr,int maxc) {
	int	i,j;
	int	tnodes,nodes,freeptr,n1,n2,total;
	int	bits;
	ushort	code=0;

	// set globals

	hdp = hd_ptr;

	// initialize all tree nodes

	for(i=0;i<maxc;i++) {
		hdp->left[i]=hdp->right[i]=-1;
		hdp->bitlen[i]=0;
		hdp->tdepth[i]=0;
		bitlen_table[i]=0;
	}

	// search histogram, find out how many nodes we have
	// add each node to translation matrix

	nodes=0;
	for(i=0;i<maxc;i++)
		if (hdp->histo[i]) {
			hdp->tmatrix[nodes] = i;
			++nodes;
		}

	tnodes=total=nodes;
	freeptr = maxc;	// first new free location

	// sort translation matrix

	qsort(hdp->tmatrix,nodes,sizeof hdp->tmatrix[0],HuffmanHistoPassCompare);

	// group least common nodes until we have only one node

	while(nodes>1) {
		n1 = hdp->tmatrix[nodes-1];
		n2 = hdp->tmatrix[nodes-2];
		hdp->left [freeptr] = n1;
		hdp->right[freeptr] = n2;
		hdp->tdepth[freeptr] = max(hdp->tdepth[n1],hdp->tdepth[n2])+1;

		hdp->histo[freeptr] = hdp->histo[n1]+hdp->histo[n2];

		// figure out where we should insert this node into sorted histogram

		i = nodes-2;	// the bottom of the new list
		while(i && hdp->histo[hdp->tmatrix[i-1]]<hdp->histo[freeptr]) --i;

		// shift rest of list down, insert new node

		nodes--;		// shorten list by 1
		memmove(hdp->tmatrix+i+1,hdp->tmatrix+i,sizeof(hdp->tmatrix[0])*(total-i));
		hdp->tmatrix[i] = freeptr;
		++freeptr,++total;
	}

	// assign bitlengths

	hdp->histo_overflow = 0;
	memset(hdp->bitlcnt,0,sizeof(HuffInt)*(MAX_LEN+1));
	HistoAssignBitlen(hdp->tmatrix[0],0);

	memcpy(bitlen_table,hdp->bitlen,sizeof(char)*maxc);
	if (hdp->histo_overflow) {
		do {
			bits = MAX_LEN-1;
			while(!hdp->bitlcnt[bits]) --bits;
			hdp->bitlcnt[bits]	--;
			hdp->bitlcnt[bits+1] += 2;
			hdp->bitlcnt[MAX_LEN]--;
			hdp->histo_overflow-=2;
		} while(hdp->histo_overflow>0);
		for(bits = MAX_LEN;bits;bits--) {
			i = hdp->bitlcnt[bits];
			while(i--) {
				do j=hdp->tmatrix[--total]; while(j>=maxc);
				bitlen_table[j] = bits;
			}
		}
	}

	// construct encoding table

	for(j=1;j<=MAX_LEN;j++) {
		for(i=0;i<maxc;i++) {
			if (bitlen_table[i]==j) {
				e_table[i] = code>>(MAX_LEN-j);

				code += (1L<<MAX_LEN)>>j;

			}
		}
	}
}

static void HistoAssignBitlen(int inode,int blen) {
	if (hdp->left[inode]>-1) {
		HistoAssignBitlen(hdp->left[inode],blen+1);
		HistoAssignBitlen(hdp->right[inode],blen+1);
		if (blen>MAX_LEN) {
			++hdp->histo_overflow;
		}
	} else {
		if (blen>MAX_LEN) {
			hdp->bitlen[inode] = blen = MAX_LEN;
			++hdp->histo_overflow;
		} else {
			hdp->bitlen[inode] = blen;
		}

		hdp->bitlcnt[blen]++;
	};
}

///////////////////////////////////////////////////////////////////////////

uchar	dist_bits_tbl[]={
	0,0,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10
};

ulong	dist_tbl[]={
	1,2,3,4,5,7,9,13,17,25,33,49,65,97,129,193,257,385,513,769,1025,1537,2049,3073,4097
};

uchar	len_bits_tbl[]={
	0,0,0,0,0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,0
};

ulong	len_tbl[]={
	3,4,5,6,7,8,9,10,11,13,15,17,19,23,27,31,35,43,51,59,67,83,84
};

ulong	histo_ldt[280];
ulong	histo_len[22];
ushort	huff_ld_encode[280];
ushort	huff_len_encode[22];
ushort	huff_ldlen_encode[16];
uchar	huff_ld_bits[280];
uchar	huff_len_bits[22];
uchar	huff_ldlen_bits[16];
struct HuffmanData hd;

unsigned long EncodeHuffman(uchar *inmem,uchar *tomem,long inbytes)
{
	int  i, c, len, r, s, last_match_length;
	unsigned long codes=0;

	unsigned short index=0x10000-F2;

	uchar *start = tomem;
	uchar *outptr = malloc(inbytes*2);
	uchar *outmem = outptr;

	/////////////

	memset(histo_ldt,0,sizeof histo_ldt);
	memset(histo_len,0,sizeof histo_len);

	codesize = printcount = 0;

	InitTree();

	s = 0;  r = N - F2;
	for (i = s; i < r; i++) text_buf[i] = 0;
	for (len = 0; len < F2; len++) {
		if (inbytes--<=0) break;
		text_buf[r + len] = *inmem++;
	}
	if ((textsize = len) == 0) {
		*(unsigned long *)start = 0;
		return 0;
	}
//	for (i = 1; i <= F2; i++) InsertNode2(r - i);
	InsertNode2(r);
	do {
		if (match_length > len) match_length = len;
		if (match_length <= THRESHOLD) {
			match_length = 1;

			*(ushort *)outptr = text_buf[r];	outptr+=2;

			++histo_ldt[text_buf[r]];
		} else {
			int code;

			match_position = 0x1000 - ((match_position - index)&0xfff);

			*(ushort *)outptr = match_position+256;	outptr+=2;
			*(uchar *)outptr = match_length;	++outptr;

			// update histogram

			code = 0;
			while(match_position>=dist_tbl[code]) ++code;

			++histo_ldt[256+code-1];

//			printf("LZ match: distance code %d ",code-1);

			code = 0;
			while(match_length>=len_tbl[code]) ++code;
			++histo_len[code-1];

//			printf("length code %d\n",code-1);
		}
		++codes;
		index+=match_length;
		last_match_length = match_length;
		for (i = 0; i < last_match_length; i++) {
			if (inbytes--<=0) break;
			DeleteNode(s);
			text_buf[s] = *inmem++;

			if (s < F2 - 1) text_buf[s + N] = c;
			s = (s + 1) & (N - 1);  r = (r + 1) & (N - 1);
			InsertNode2(r);
		}
		textsize += i;
		while (i++ < last_match_length) {
			DeleteNode(s);
			s = (s + 1) & (N - 1);  r = (r + 1) & (N - 1);
			if (--len) InsertNode2(r);
		}
	} while (len > 0);

	// check to see if either of the tables only has a single code
	// if so, add code 0 so the tables won't be all zero

	for(i=0;i<280;i++) if (histo_ldt[i]) break;
	if (i<280) {
		for(++i;i<280;i++) if (histo_ldt[i]) break;
		if (i>=280) {
			if (histo_ldt[0]) histo_ldt[1]=1;
			else histo_ldt[0]=1;
		}
	}

	for(i=0;i<22;i++) if (histo_len[i]) break;
	if (i<22) {
		for(++i;i<22;i++) if (histo_len[i]) break;
		if (i>=22) {
			if (histo_len[0]) histo_len[1]=1;
			else histo_len[0]=1;
		}
	}

	// construct huffman encoding tables for literal/distance and length tables

	memcpy(hd.histo,histo_len,sizeof(ulong)*22);
	HuffmanBuildTree(huff_len_encode,huff_len_bits,&hd,22);
	memcpy(hd.histo,histo_ldt,sizeof(ulong)*280);
	HuffmanBuildTree(huff_ld_encode,huff_ld_bits,&hd,280);
	for(i=0;i<16;i++) hd.histo[i] = 0;
	for(i=0;i<280;i++) hd.histo[huff_ld_bits[i]]++;
	HuffmanBuildTree(huff_ldlen_encode,huff_ldlen_bits,&hd,16);

	// write out trees

	for(i=0;i<16;i+=2)
		*tomem++ = huff_ldlen_bits[i] | (huff_ldlen_bits[i+1]<<4);

	for(i=0;i<22;i+=2)
		*tomem++ = huff_len_bits[i] | (huff_len_bits[i+1]<<4);

	{
		ulong	accum;
		uchar	accum_bits = 0;
		uchar	*src;
		int	c,code;

		// write out main l/d tree

		for(i=0;i<280;i++) {
			c = huff_ld_bits[i];
//			printf("L/D code %3d: %d bits, encoding %08lx(%d)\n",i,c,huff_ldlen_encode[c],huff_ldlen_bits[c]);

			if (!huff_ldlen_bits[c])
				Error("Zero bits for L/D bitlen %ld on code %d (occurs %ld times)\n",c,i,histo_ldt[i]);

			accum <<= huff_ldlen_bits[c];
			accum_bits += huff_ldlen_bits[c];
			accum |= huff_ldlen_encode[c];

			if (accum_bits >= 16) {
				*tomem++ = accum >> (accum_bits-8);
				*tomem++ = accum >> (accum_bits-16);
				accum_bits -= 16;
			}
		}

		// write out data

		src = outmem;
		while(src<outptr) {
			c = *(ushort *)src; src+=2;

			if (c < 256) {	// literal

//			if (!huff_ld_bits[c]) Error("Zero bits for code %d (literal)\n",c);

				accum <<= huff_ld_bits[c];
				accum_bits += huff_ld_bits[c];
				accum |= huff_ld_encode[c];

//				printf("Literal %d\n",c);

			} else {	// LZ match


				c-=256;

				code = 0;
				while(c>=dist_tbl[code]) ++code;
				--code;

//				printf("LZ match: %08lx,%ld (code %ld)\n",c,*src,code);

//			if (!huff_ld_bits[code+256]) Error("Zero bits for code %d (distance %ld), hs=%ld\n",code,c,histo_ldt[code+256]);

				accum <<= huff_ld_bits[code+256];
				accum_bits += huff_ld_bits[code+256];
				accum |= huff_ld_encode[code+256];

				if (accum_bits >= 16) {
					*tomem++ = accum >> (accum_bits-8);
					*tomem++ = accum >> (accum_bits-16);
					accum_bits -= 16;
				}

				// output extra bits for distance

				accum <<= dist_bits_tbl[code];
				accum_bits += dist_bits_tbl[code];
				accum |= (c-dist_tbl[code]);

				if (accum_bits >= 16) {
					*tomem++ = accum >> (accum_bits-8);
					*tomem++ = accum >> (accum_bits-16);
					accum_bits -= 16;
				}

				// output length code

				c = *src++;

				code = 0;
				while(c>=len_tbl[code]) ++code;
				--code;

//			if (!huff_len_bits[code]) Error("Zero bits for code %d (distance)\n",code);

				accum <<= huff_len_bits[code];
				accum_bits += huff_len_bits[code];
				accum |= huff_len_encode[code];

				if (accum_bits >= 16) {
					*tomem++ = accum >> (accum_bits-8);
					*tomem++ = accum >> (accum_bits-16);
					accum_bits -= 16;
				}

				// output extra bits for length

				accum <<= len_bits_tbl[code];
				accum_bits += len_bits_tbl[code];
				accum |= (c - len_tbl[code]);
			}


			if (accum_bits >= 16) {
				*tomem++ = accum >> (accum_bits-8);
				*tomem++ = accum >> (accum_bits-16);
				accum_bits -= 16;
			}
		}

		// flush

		if (accum_bits) {
			accum <<= (16-accum_bits);
			*tomem++ = accum>>8;
			*tomem++ = accum;
		}
	}

//	printf("%ld -> %ld\n", textsize,tomem-start);

	free(outmem);

	return tomem - start;
}

//////////////////////////////////////////////

struct LEHeader {
	ushort	magic_le;	// 00 - should be 'EL' (LE backwards)
	boolean	byte_order;	// 02 - byte order
	boolean	word_order;	// 03 - word order
	ulong	format_level;	// 04 - format level
	ushort	cpu_type;	// 08 - CPU type req'd
	ushort	os_type;	// 0a - OS type req'd
	ulong	module_version;	// 0c - version of module
	ulong	module_flags;	// 10 - module flags
	ulong	module_pages;	// 14 - # of pages in this module
	ulong	EIP_object;	// 18 - object EIP offset refers to
	ulong	EIP_offset;	// 1c - EIP offset within that object
	ulong	ESP_object;	// 20 - object ESP offset refers to
	ulong	ESP_offset;	// 24 - ESP offset within that object
	ulong	page_size;	// 28 - page size, in bytes
	ulong	final_page;	// 2c - # of bytes in final 4k image page, 0->4095
	ulong	relosize;	// 30 - length of relocation section
	ulong	relochk;	// 34 - relocation checksum (not used)
	ulong	loadersize;	// 38 - loader section size
	ulong	loaderchk;	// 3c - loader section checksum
	ulong	object_offset;	// 40 - offset to object table - from LE header
	ulong	object_cnt;	// 44 - # of objects in object table
	ulong	objpage_offset;	// 48 - offset to object page table
	ulong	objipg_offset;	// 4c - offset to object iterated page table
	ulong	resrc_offset;	// 50 - offset to resource table
	ulong	resrc_entries;	// 54 - # of entries in resource table
	ulong	name_offset;	// 58 - offset to name
	ulong	entry_tbl;	// 5c - entry name table offset
	ulong	dirctv_off;	// 60 - directive table offset
	ulong	dirctv_cnt;	// 64 - directive table count
	ulong	relooffs_tbl;	// 68 - relocation page table offset - offsets to relocs for each page
	ulong	relopage_tbl;	// 6c - relocation page number table offset
	ulong	impname_tbl;	// 70 - import module name table offset
	ulong	impname_cnt;	// 74 - # of entries in import module name table
	ulong	impproc_tbl;	// 78 - import procedure table offset
	ulong	pagechk_tbl;	// 7c - per-page checksum table offset
	ulong	image_start;	// 80 - absolute file offset of image
	ulong	preload_pages;	// 84 - # of pages to preload
	ulong	nonres_tbl;	// 88 - nonresident name table offset
	ulong	nonres_len;	// 8c - nonresident name table length
	ulong	nonres_chk;	// 90 - nonresident name table checksum
	ulong	autods_obj;	// 94 - auto-ds object -- unused for 32-bit exe's
	ulong	debug_off;	// 98 - offset to debug info
	ulong	debug_len;	// 9c - length of debug info
	ulong	inst_preload;	// a0 - # of instance data pages in preload section
	ulong	inst_demand;	// a4 - # of instance data pages in demand section
	ulong	heapsize;	// a8 - heap size to add to auto-ds segment - unused for 32-bit exe's
};

struct LEModule {
	struct LEModule *next;
	struct LEHeader leh;
};

struct	LEObject {
	ulong	size;		// size - what more can I say?
	ulong	rbase_addr;	// relocation base address - address obj currently relocated to
	ulong	flags;		// flags
	ulong	first_page;	// first page # in page number table
	ulong	pagecnt;	// # of entries in page # table for this object
	ulong	segment;	// really a reserved, but internal pointer for me
};

#define	RB_SIZE (4096)

struct	RelocBlock {
	struct RelocBlock *next;
	ulong	numrelocs;
	ulong	relocs[RB_SIZE];
};

void	Error(char *s, ...) {
	va_list val;

	fputs("le2ve: ",stdout);
	va_start(val,s);
	vprintf(s,val);
	va_end(val);
	puts(".");
	exit(5);
}

void	Status(char *s, ...) {
	va_list val;

	fputs("le2ve: ",stdout);
	va_start(val,s);
	vprintf(s,val);
	va_end(val);
	puts(".");
}

void	Warning(char *s, ...) {
	va_list val;

	fputs("le2ve: ",stdout);
	va_start(val,s);
	vprintf(s,val);
	va_end(val);
	puts("!");
}

char	err_mem[]="not enough memory";
char	err_mem_iname[]="not enough memory for import name table";
char	err_mem_iproc[]="not enough memory for import procedure table";
char	err_mem_ldata[]="not enough memory for library data";
char	err_mem_reloc[]="not enough memory for relocs";
char	err_open_file[]="can't open file";
char	err_read_file[]="can't read file";
char	err_write_file[]="can't write file";
char	err_seek_file[]="can't seek file";
char	err_close_file[]="can't close file";
char	err_read_stub[]="error reading VELoader stub";

void	AddReloc(struct RelocBlock **rbp,ulong addr) {
	struct RelocBlock *rb,*rbt;

	rb = *rbp;

	if (rb ? rb->numrelocs == RB_SIZE : TRUE) {
		if (!(rbt = malloc(sizeof(struct RelocBlock)))) Error(err_mem_reloc);
		rbt->next	= rb;
		rbt->numrelocs	= 0;
		rb = rbt;
	}

	rb->relocs[rb->numrelocs++] = addr;

	*rbp = rb;
}

ulong	*RelocChainToList(struct RelocBlock *rb) {
	struct RelocBlock *rbt;
	ulong	len=0,tlen;
	ulong	*l;

	rbt=rb;
	while(rbt) {
		len += rbt->numrelocs;
		rbt=rbt->next;
	}

	tlen = len++;

	if (!(l = malloc(sizeof(ulong)*len))) return NULL;

	while(rbt=rb) {
		len -= rbt->numrelocs;
		memcpy(l+len,rbt->relocs,rbt->numrelocs*sizeof(ulong));
		rb=rbt->next;
		free(rbt);
	}

	l[0] = tlen;

	return l;
}

uchar *	RelocProcess(ulong *r,uchar type,ulong ln) {
	ulong	len=*r++;
	ulong	rb=0;
	ulong	last,now;
	long	diff;
	uchar	*m,*t;

	if (!(m = t = malloc(len*6+15))) return NULL;

	t+=4;

	while(len) {
		if (!rb) {
			rb = min(len,65536);
			t[0] = type;
			*(ushort *)(t+1) = rb;
			*(ulong *)(t+3) = last = *r++;
			t += 7;
			--rb;
			--len;
			continue;
		}
		--len;

		now = *r++;
		diff = now-last-ln;
		last = now;

		if (diff<0) Error("overlapping relocs!");

#if 0
		if	(diff & 0xf0000000L) *t++ = 0x80 + ((diff>>28) & 0x7f);
		if	(diff & 0x0fe00000L) *t++ = 0x80 + ((diff>>21) & 0x7f);
		if	(diff & 0x001fc000L) *t++ = 0x80 + ((diff>>14) & 0x7f);
		if	(diff & 0x00003f80L) *t++ = 0x80 + ((diff>> 7) & 0x7f);
#else
		if	(diff & 0xf0000000L) *t++ = 0x80 + ((diff>>28) & 0x7f);
		if	(diff & 0xffe00000L) *t++ = 0x80 + ((diff>>21) & 0x7f);
		if	(diff & 0xffffc000L) *t++ = 0x80 + ((diff>>14) & 0x7f);
		if	(diff & 0xffffff80L) *t++ = 0x80 + ((diff>> 7) & 0x7f);
#endif
		*t++ =	 diff & 0x0000007fL;

		--rb;
	}

	realloc(m,t-m);
	*(ulong *)m = t-m-4;

	return m;
}

int	sortulong(const ulong *_1,const ulong *_2) {
	if (*_1>*_2) return 1;
	else if (*_1<*_2) return -1;
	else return 0;
}

char instructions[]=
	"\nLE2VE V1.21, Copyright (C) 1997 by Avery Lee.  All Rights Reserved.\n"
	"%s [options] <LE executable> <VE executable>\n"
	"\n"
	"-i0"	"\t"	"Image: No compression\n"
	"-i1"	"\t"	"Image: LZSS compression\n"
	"-i2"	"\t"	"Image: LZH compression\n"
	"-i9"	"\t"	"Image: try all compression methods (default)\n"
	"-r0"	"\t"	"Relocs: No compression\n"
	"-r1"	"\t"	"Relocs: LZSS compression\n"
	"-r2"	"\t"	"Relocs: LZH compressionn"
	"-r9"	"\t"	"Relocs: try all compression methods (default)\n"
	"-s"	"\t"	"Prepend VELOADER.EXE stub to VE executable\n"
	"-S"	"\t"	"Prepend VELOD4GW.EXE stub to VE executable\n"
	"-Q"	"\t"	"Modify stub for quiet decompression\n"
	"\n"
	"LE2VE is freely distributable under the terms of the GNU General\n"
	"Public License Version 2 or greater; see the VE archive for more\n"
	"details.\n\n"
;

char *argv0;

void	ReadTheStupidManual(void) {
	printf(instructions,argv0);
	exit(5);
}

enum {
	COMP_NONE	= 0,
	COMP_LZSS	= 1,
	COMP_LZH	= 2,
	COMP_BEST	= 9
};

int main(int argc,char **argv) {
	// allocated structures

	FILE	*exe;
	FILE	*out;
	struct LEModule *lem		=NULL;
	void	*progmem		=NULL;
	uchar	*relocs			=NULL;
	uchar	*import_module_table	=NULL;
	uchar	*import_proc_table	=NULL;
	uchar	*librarymem		=NULL;

	struct	RelocBlock	*reloc32	= NULL;
	struct	RelocBlock	*reloc48	= NULL;
	struct	RelocBlock	*reloc16	= NULL;
	ulong	*reloc32l;
	ulong	*reloc16l;
	ulong	*reloc48l;
	uchar	*reloc32p;
	uchar	*reloc16p;
	uchar	*reloc48p;

	uchar	*compmem,*compmemr,*compmeml,*compmeme;
	uchar	*relocmem,*exportmem;

	// locals

	ulong	*reloffstbl		=NULL;
	uchar	*reldatatbl		=NULL;

	struct LEHeader leh;
	struct LEObject *obj;
	ulong LEstart=0,MZstart;
	ulong totalsize;
	int i;
	struct VExeHeader vpeh;

	ulong	imagesize,relocsize;
	ulong	highest_reloc_point=0;
	ulong	librarysize		= 8;
	ulong	exportsize		= 0;
	ulong	packimage_size,packrelocs_size,packlibrary_size,packexport_size;

	int	image_compression	= COMP_BEST;
	int	relocs_compression	= COMP_BEST;
	int	library_compression	= COMP_BEST;
	int	export_compression	= COMP_BEST;
	boolean	quiet_decompression	= FALSE;

	boolean	prepend=FALSE;
	uchar	prepend_path[256];
	int	dos4gwpro=0;

	int	library_name_count	= 0;
	int	library_module_count	= 0;

	////////////////////////////////////////////////////////

	argv0 = argv[0];

	// scan for options

	while(argc>1) {
		char *s;

		s = argv[1];

		if (*s=='-') {
			switch(s[1]) {
			case 'i':	// image comp
				if ((s[2]<'0' || s[2]>'2') && s[2]!='9') ReadTheStupidManual();
				image_compression = s[2]-'0';
				break;
			case 'r':	// relocs comp
				if ((s[2]<'0' || s[2]>'2') && s[2]!='9') ReadTheStupidManual();
				relocs_compression = s[2]-'0';
				break;
			case 's':	// VELOADER.EXE
			case 'S':	// VELOD4GW.EXE
				{
					char *slash=NULL;
					char *t;

					t = argv0;
					while(*t++) if (t[-1]==':' || t[-1]=='\\') slash=t;

					if (!slash) Error("can't find path to %s",s[1]=='s' ? "VELoader.EXE" : "VELod4GW.EXE");

					memcpy(prepend_path,argv0,slash-argv0);
					strcpy(prepend_path+(slash-argv0),s[1]=='s' ? "VELoader.EXE" : "VELod4GW.EXE");
					prepend = TRUE;
				}
				break;
			case 'Q':
				quiet_decompression = TRUE;
				break;
			default:
				ReadTheStupidManual();
				break;
			}
		} else break;

		++argv,--argc;
	}

	if (3<argc) ReadTheStupidManual();

	// open the file

	if (!(exe=fopen(argv[1],"rb"))) Error(err_open_file);

	// determine start of LE program from DOS stub EXE header

	do {
		{
			ushort tag,sectors,bytes;

			if (fseek(exe,LEstart,SEEK_SET)) Error(err_seek_file);

			if (1!=fread(&tag,sizeof tag,1,exe)) Error(err_read_file);

			if (tag=='ZM') {
				if (1!=fread(&bytes,sizeof bytes,1,exe)) Error(err_read_file);
				if (1!=fread(&sectors,sizeof sectors,1,exe)) Error(err_read_file);

				if (bytes) --sectors;

				MZstart = LEstart;
				LEstart += (sectors<<9)+bytes;
			}
		}

		Status("looking for LE at %08lx",LEstart);

		// read area, look for LE header

		if (fseek(exe,LEstart-16,SEEK_SET)) Error(err_seek_file);
		if (1!=fread(&leh,sizeof leh,1,exe)) Error(err_read_file);

		// search for LE marker

		{
			uchar *lhp = (uchar *)&leh;

			for(i=0;i<64;i++,lhp++) {
				if (((struct LEHeader *)lhp)->magic_le == 'EL') break;

				// the BW tag is internal to DOS/4GW Professional

				if (((struct LEHeader *)lhp)->magic_le == 'WB') break;
			}

			if (i>=64) Error("LE tag search failed");

			LEstart = LEstart - 16 + i;
		}

		// read in LE header

		if (fseek(exe,LEstart,SEEK_SET)) Error(err_seek_file);
		if (1!=fread(&leh,sizeof leh,1,exe)) Error(err_read_file);

		if (leh.magic_le!='EL' && leh.magic_le != 'WB') Error("LE tag not found");		// not LE

		if (leh.magic_le == 'WB') {
			LEstart += leh.ESP_object;	// 0x20 from start; not really ESP
			dos4gwpro = 1;
		}
	} while(dos4gwpro--);

	// read in object table

	Status("reading in object table");

	if (!(obj = malloc(sizeof(struct LEObject)*leh.object_cnt))) goto abort;

	if (fseek(exe,LEstart + leh.object_offset,SEEK_SET)) goto abort;

	if (1!=fread(obj,sizeof(struct LEObject)*leh.object_cnt,1,exe)) goto abort;

	{
		ulong progptr = 0;

		totalsize = 0;

		for(i=0;i<leh.object_cnt;i++) {
			Status("Object %d: %08lx bytes",i+1,obj[i].size);

			totalsize += (obj[i].size+0xfff)&0xfffff000;
			obj[i].segment = progptr;
			progptr += (obj[i].size+0xfff)&0xfffff000;
		}
	}

	// allocate memory for executable

	imagesize = leh.module_pages<<12;
	if (leh.final_page) imagesize-=4096;
	imagesize += leh.final_page;

	if (!(progmem=malloc(totalsize))) goto abort;
	memset(progmem,0,totalsize);

	// allocate memory for relocs

	if (!(relocs = malloc(leh.relosize))) goto abort;

	reloffstbl = (ulong *)relocs;
	reldatatbl = (uchar *)(relocs + (leh.module_pages<<2) + 4);

	// read program image in

	Status("reading program image");

	if (fseek(exe,MZstart + leh.image_start,SEEK_SET)) goto abort;
	if (1!=fread(progmem,imagesize,1,exe)) goto abort;

	// read relocs in

	Status("reading relocs (%ld bytes)",leh.relosize);

	if (fseek(exe,LEstart + leh.relooffs_tbl,SEEK_SET)) Error(err_seek_file);
	if (1!=fread(reloffstbl,(leh.module_pages<<2)+4,1,exe)) Error(err_read_file);

	if (fseek(exe,LEstart + leh.relopage_tbl,SEEK_SET)) Error(err_seek_file);
	if (1!=fread(reldatatbl,leh.relosize-((leh.module_pages<<2)+4),1,exe)) Error(err_read_file);

	// read import module name table in

	if (leh.impname_tbl != leh.impproc_tbl) {
		char	*p;
		int	j;

		Status("reading import module name table (%ld bytes)",leh.impproc_tbl - leh.impname_tbl);

		if (!(import_module_table = malloc(leh.impproc_tbl - leh.impname_tbl))) Error(err_mem_iname);
		if (fseek(exe,LEstart + leh.impname_tbl,SEEK_SET)) Error(err_seek_file);
		if (1!=fread(import_module_table,leh.impproc_tbl - leh.impname_tbl,1,exe)) Error(err_read_file);

		// figure out how much space is needed for the module name table

		p = import_module_table;
		for(i=0;i<leh.impname_cnt;i++) {
			j = *p++;
			if (j) {
				librarysize += j+1;
				++library_module_count;
			}
		}

	}

	// read import procedure name table

	if (leh.impproc_tbl) {
		ulong	ipt_size;
		char	*p;
		int	j;

		ipt_size = leh.relooffs_tbl + leh.relosize - leh.impname_tbl;

		Status("reading import procedure name table (%ld bytes)",ipt_size);

		if (!(import_proc_table = malloc(ipt_size))) Error(err_mem_iproc);
		if (fseek(exe,LEstart + leh.impproc_tbl,SEEK_SET)) Error(err_seek_file);
		if (1!=fread(import_proc_table,ipt_size,1,exe)) Error(err_read_file);

		// display names

		p = import_proc_table;
		while(p < import_proc_table + ipt_size) {
			j = *p++;
			if (j) {
				librarysize += j+1;
				++library_name_count;
			}
		}

	}

	// process exports
#if 0
	if (leh.
	Status("reading export table");
#endif
	exportsize = packexport_size = 0;

	// close file

	fclose(exe);

#if 0
	// process library data

	if (librarysize > 8) {
		char	*ipt_limit;
		char	*s,*t;
		int	j;

		Status("processing library data");

		if (!(librarymem = malloc(librarysize))) Error(err_mem_ldata);

		*(ulong *)librarymem = library_module_count;

		s = import_module_table;
		t = librarymem + 4;
		for(i=0;i<leh.impname_cnt;i++) {
			j = *s++;
			if (j) {
				*t++ = j;
				while(j--) *t++ = toupper(*s++);
			}
		}

		*(ulong *)t = library_name_count; t+=4;
		s = import_proc_table;
		ipt_limit = s + leh.relooffs_tbl + leh.relosize - leh.impname_tbl;

		while(s < ipt_limit) {
			j = *s++;
			if (j) {
				*t++ = j;
				while(j--) *t++ = toupper(*s++);
			}
		}

		// compress library data

		if (library_compression) {
			printf("le2ve: compressing library data..."); fflush(stdout);

			switch(library_compression) {
			case COMP_LZH:
				if (!(compmeml = malloc(librarysize*2))) Error(err_mem);
				packlibrary_size = EncodeHuffman(librarymem,compmeml,librarysize);
				break;
			case COMP_LZSS:
				if (!(compmeml = malloc(librarysize + (librarysize>>3)))) Error(err_mem);
				packlibrary_size = Encode(librarymem,compmeml,librarysize);
				break;
			}
		}

	} else
#endif
		librarysize = packlibrary_size = 0;

	// process relocs

	Status("processing relocs");

	{
		uchar	*reloptr	= reldatatbl;
		ulong	*reloffst	= reloffstbl;
		uchar	*relonxt	= reldatatbl;
		uchar	*relotop	= relocs + leh.relosize;
		uchar	*memptr;
		ulong	offset;
		uchar	*progptr = progmem;
		uchar	flags;
		ushort	object;
		int	page;
		int	source_count;
		ulong	reloc_size;

		ulong	value;

		++reloffst;

		relonxt += *reloffst++;
		for(page=0;page<leh.module_pages;page++) {
			while(reloptr < relonxt) {
				uchar	type;

				type = *reloptr++;

				if (type & 0xd0)
					Warning("reloc type flags %04x ignored",type & 0xd0);
#if 0
				Status("%02x %02x %02x %02x %02x %02x %02x %02x",type,reloptr[0],reloptr[1],reloptr[2],reloptr[3],reloptr[4],reloptr[5],
					reloptr[6]);
#endif
				switch(type & 0x0f) {
				case 0x02:
				case 0x05:
				case 0x06:
				case 0x07:
				case 0x08:
					flags = *reloptr++;

					// 03h = Fixup target type mask (00h, internal only supported)
					// 04h = Additive Fixup Flag
					// 10h = 32-bit Target Offset Flag (supported)
					// 20h = 32-bit Additive Fixup Flag (ignored)
					// 40h = 16-bit Object Number/Module Ordinal Flag (partially supported)
					// 80h = 8-bit Ordinal Flag (ignored)

					if (flags & 0x0f) {
//						puts("Error in LE file");
//						goto abort;

						Error("LE relocation has unknown flags: %02x",flags);
					}

					if (type & 0x20) {
						source_count = *reloptr++;
					} else {
						source_count = 1;
						offset = *(ushort *)reloptr;	reloptr+=2;
						memptr = progptr + (offset & 0xFFF);
					}

					if (flags & 0x40) {
						object = *(ushort *)reloptr;
						reloptr += 2;
					} else
						object = *reloptr++;

					if ((type&0x0f) == 0x08) {
						if (flags & 0x10) {
							value = *(long *)reloptr;
							reloptr += 4;
						} else {
							value = *(short *)reloptr;
							reloptr += 2;
						}
						value += progptr-memptr;
					} else if ((type & 0x0f) != 0x02) {
						value = (ulong)obj[object-1].segment;

						if (flags & 0x10) {
							value += *(ulong *)reloptr;
							reloptr += 4;
						} else {
							value += *(ushort *)reloptr;
							reloptr += 2;
						}
					}

					do {
						if (type & 0x20) {
							offset = *(ushort *)reloptr;	reloptr+=2;
							memptr = progptr + (offset & 0xFFF);
						}

						if (offset < 0x1000) {
							// this relocation has not already been done in the previous page

							switch(type & 0x0f) {
							case 0x02:	// 16-bit selector fixup
								*(ushort *)memptr = !(obj[object-1].flags & 0x4000);
								AddReloc(&reloc16,memptr-progmem);
								reloc_size = 2;
								break;
							case 0x05:	// 16-bit offset fixup
								Warning("16-bit offset reloc (0x05) found");
								break;
							case 0x06:	// 16:32 far pointer fixup
								*(ulong *)memptr = value;
								*(ushort *)(memptr+4) = !(obj[object-1].flags & 0x4000);
								AddReloc(&reloc48,memptr-progmem);
								reloc_size = 6;
								break;
							case 0x07:	// 32-bit offset
								*(ulong *)memptr = value;
								AddReloc(&reloc32,memptr-progmem);
								reloc_size = 4;
								break;
							case 0x08:	// 32-bit self relative offset fixup
								*(ulong *)memptr = value;
								AddReloc(&reloc32,memptr-progmem);
								reloc_size = 4;
								break;
							}
						}
						if (highest_reloc_point < (memptr-progmem)+reloc_size)
							highest_reloc_point = (memptr-progmem)+reloc_size;
					} while(--source_count);

					break;
				default:
					Warning("unknown reloc type: %02x",type);
				}
			}
			relonxt = reldatatbl + *reloffst++;
			progptr+= 4096;
			continue;
		}
	}

	//
	//
	// Watcom's linker attempts to save space in LE executables by trimming
	// off zeroes from the end of the data segment, effectively making them
	// BSS.  BSS _must_ be zeroed, so this isn't normally a problem. 
	//
	// HOWEVER, the VE format requires that any relocated pointers hold
	// offsets in the image location.  This is not true for LE, because LE
	// puts the offset data in the relocation data block.  Thus we must
	// examine the reloc stream and see if any relocs point beyond the
	// image; if so, extend the image size to hold the VE offset data.
	//
	//

	if (highest_reloc_point > imagesize) {
		Warning("warning, reported size is %08lx, true size is at least %08lx",imagesize,highest_reloc_point);
		imagesize = highest_reloc_point;
	}

	//
	// free relocs
	//

	free(relocs);

	//
	// allocate memory for reloc lists
	//

	if (!(reloc32l = RelocChainToList(reloc32))) Error(err_mem_reloc);
	if (!(reloc48l = RelocChainToList(reloc48))) Error(err_mem_reloc);
	if (!(reloc16l = RelocChainToList(reloc16))) Error(err_mem_reloc);

	Status("%ld 32-bit relocs, %ld 48-bit relocs, %ld 16-bit relocs",
		reloc32l[0],
		reloc48l[0],
		reloc16l[0]);

	//
	// sort relocation lists
	//

	Status("sorting relocations");

	qsort(reloc32l+1,reloc32l[0],sizeof(ulong),sortulong);
	qsort(reloc48l+1,reloc48l[0],sizeof(ulong),sortulong);
	qsort(reloc16l+1,reloc16l[0],sizeof(ulong),sortulong);

	if (!(reloc32p = RelocProcess(reloc32l,VEXE_RELOC_32BIT,4))) Error(err_mem_reloc);
	if (!(reloc48p = RelocProcess(reloc48l,VEXE_RELOC_48BIT,6))) Error(err_mem_reloc);
	if (!(reloc16p = RelocProcess(reloc16l,VEXE_RELOC_16BIT,2))) Error(err_mem_reloc);

	Status("%ld bytes for 32-bit, %ld bytes for 48-bit, %ld bytes for 16-bit"
		,*(ulong *)reloc32p
		,*(ulong *)reloc48p
		,*(ulong *)reloc16p
	);

	if (image_compression) {
		printf("le2ve: compressing image..."); fflush(stdout);


		switch(image_compression) {
		case COMP_BEST:
			{
			ulong packsize_lzss;
			ulong packsize_lzh;

			if (!(compmem = malloc(imagesize*3+(imagesize>>3)))) Error(err_mem);	// wasteful as hell

			packsize_lzh  = EncodeHuffman(progmem,compmem,imagesize);
			packsize_lzss = Encode(progmem,compmem+imagesize*2,imagesize);

			if (packsize_lzh < packsize_lzss) {
				image_compression = COMP_LZH;
				packimage_size = packsize_lzh;
			} else {
				image_compression = COMP_LZSS;
				packimage_size = packsize_lzss;
				memmove(compmem,compmem+imagesize*2,packimage_size);
			}

			} break;
		case COMP_LZH:
			if (!(compmem = malloc(imagesize*2))) Error(err_mem);	// ought to be enough
			packimage_size = EncodeHuffman(progmem,compmem,imagesize);
			break;
		case COMP_LZSS:
			if (!(compmem = malloc(imagesize+(imagesize>>3)))) Error(err_mem);	// ought to be enough
			packimage_size = Encode(progmem,compmem,imagesize);
			break;
		}

		// if we expanded the stupid image, use uncompressed mode

		if (packimage_size > imagesize) {
			packimage_size = imagesize;
			printf("%ld (couldn't compress)\n",imagesize);
		} else {
			printf("%ld -> %ld (%s)\n",imagesize,packimage_size,image_compression==COMP_LZH ? "LZH" : "LZSS");
		}

	} else packimage_size = imagesize;

	// compute total relocs size

	relocsize = *(ulong *)reloc32p + *(ulong *)reloc16p + *(ulong *)reloc48p + 1;

	// allocate memory for all relocs & copy

	if (!(relocmem = malloc(relocsize))) Error(err_mem_reloc);

	memcpy(relocmem,reloc32p+4,*(ulong *)reloc32p);
	memcpy(relocmem + *(ulong *)reloc32p,reloc16p+4,*(ulong *)reloc16p);
	memcpy(relocmem + *(ulong *)reloc32p + *(ulong*)reloc16p,reloc48p+4,*(ulong *)reloc48p);
	relocmem[relocsize-1] = VEXE_RELOC_END;

	// pack relocs

	if (relocs_compression) {
		printf("le2ve: compressing relocs..."); fflush(stdout);
		switch(relocs_compression) {
		case COMP_BEST:
			{
			ulong packsize_lzss;
			ulong packsize_lzh;

			if (!(compmemr = malloc(relocsize*3+(relocsize>>3)))) Error(err_mem);	// wasteful as hell

			packsize_lzh  = EncodeHuffman(relocmem,compmemr,relocsize);
			packsize_lzss = Encode(relocmem,compmemr+relocsize*2,relocsize);

			if (packsize_lzh < packsize_lzss) {
				relocs_compression = COMP_LZH;
				packrelocs_size = packsize_lzh;
			} else {
				relocs_compression = COMP_LZSS;
				packrelocs_size = packsize_lzss;
				memmove(compmemr,compmemr+relocsize*2,packrelocs_size);
			}

			} break;
		case COMP_LZH:
			if (!(compmemr = malloc(relocsize*2))) Error(err_mem_reloc);
			packrelocs_size = EncodeHuffman(relocmem,compmemr,relocsize);
			break;
		case COMP_LZSS:
			if (!(compmemr = malloc(relocsize+(relocsize>>3)))) Error(err_mem_reloc);
			packrelocs_size = Encode(relocmem,compmemr,relocsize);
			break;
		}
		if (packrelocs_size > relocsize) {
			packrelocs_size = relocsize;
			printf("%ld (couldn't compress)\n",relocsize);
		} else {
			printf("%ld -> %ld (%s)\n",relocsize,packrelocs_size,relocs_compression==COMP_LZH ? "LZH" : "LZSS");
		}
	} else packrelocs_size = relocsize;

	// open output file(s)

  	vpeh.ID			= VEXE_ID;
	vpeh.version		= VEXE_VERSION;
	vpeh.image_offset	= sizeof vpeh;
	vpeh.relocs_offset	= vpeh.image_offset + packimage_size;
	vpeh.library_offset	= vpeh.library_offset + packrelocs_size;
	vpeh.export_offset	= vpeh.export_offset + packexport_size;
	vpeh.image_length	= packimage_size;
	vpeh.relocs_length	= packrelocs_size;
	vpeh.library_length	= packlibrary_size;
	vpeh.export_length	= packexport_size;
	vpeh.uc_image_length	= imagesize;
	vpeh.uc_relocs_length	= relocsize;
	vpeh.bss_length		= totalsize - imagesize;
	vpeh.flags		= NULL;
	vpeh.entry_point	= ((obj[leh.EIP_object-1].first_page-1)<<12)+leh.EIP_offset;
	vpeh.stack_point	= ((obj[leh.ESP_object-1].first_page-1)<<12)+leh.ESP_offset;
	vpeh.resource_offset	= 0;
	vpeh.image_comp		= packimage_size<imagesize ? image_compression : VEXE_COMP_NONE;
	vpeh.relocs_comp	= packrelocs_size<relocsize ? relocs_compression : VEXE_COMP_NONE;
	vpeh.library_comp	= packlibrary_size<librarysize ? library_compression : VEXE_COMP_NONE;
	vpeh.export_comp	= packexport_size<exportsize ? export_compression : VEXE_COMP_NONE;

	if (!(out=fopen(argv[2],"wb"))) Error(err_open_file);

	if (prepend) {
		FILE *f;
		char *mem;
		int actual;

		if (!(mem = malloc(16384))) Error(err_mem);
		if (!(f=fopen(prepend_path,"rb"))) Error("error opening VELoader stub [%s]",prepend_path);

		if (quiet_decompression) {
			Status("patching stub for quiet decompression");
			if (1!=fread(mem,0x18,1,f)) Error(err_read_stub);

			// naughty, naughty....
			//
			// patch the checksum word of the 16-bit stub EXE so that
			// bit 0 is set

			*(ushort *)(mem + 0x12) |= 0x0001;

			if (1!=fwrite(mem,0x18,1,out)) Error(err_write_file);
		}

		Status("copying stub");

		while(actual = fread(mem,1,16384,f))
			if (1!=fwrite(mem,actual,1,out)) Error(err_write_file);

		if (actual<0) Error(err_read_stub);

		fclose(f);
		free(mem);
	}

	// write out values

	if (1!=fwrite(&vpeh,sizeof vpeh,1,out)) Error(err_write_file);

	if (vpeh.image_comp == VEXE_COMP_NONE) {
		if (1!=fwrite(progmem,imagesize,1,out)) Error(err_write_file);
	} else {
		if (1!=fwrite(compmem,packimage_size,1,out)) Error(err_write_file);
	}

	if (vpeh.relocs_comp == VEXE_COMP_NONE) {
		if (*(ulong *)reloc32p)	if (1!=fwrite(reloc32p+4,*(ulong *)reloc32p,1,out)) Error(err_write_file);
		if (*(ulong *)reloc48p)	if (1!=fwrite(reloc48p+4,*(ulong *)reloc48p,1,out)) Error(err_write_file);
		if (*(ulong *)reloc16p)	if (1!=fwrite(reloc16p+4,*(ulong *)reloc16p,1,out)) Error(err_write_file);
		if (EOF==fputc(0,out)) Error(err_write_file);
	} else {
		if (1!=fwrite(compmemr,packrelocs_size,1,out)) Error(err_write_file);
	}

	if (packlibrary_size) {
		if (vpeh.library_comp == VEXE_COMP_NONE) {
			if (1!=fwrite(librarymem,librarysize,1,out)) Error(err_write_file);
		} else {
			if (1!=fwrite(compmeml,packlibrary_size,1,out)) Error(err_write_file);
		}
	}

	if (packexport_size) {
		if (vpeh.export_comp == VEXE_COMP_NONE) {
			if (1!=fwrite(exportmem,exportsize,1,out)) Error(err_write_file);
		} else {
			if (1!=fwrite(compmeme,packexport_size,1,out)) Error(err_write_file);
		}
	}

	if (fclose(out)) Error(err_close_file);

	{
		ulong	iratio,rratio;

		if (packimage_size > imagesize)
			iratio = 100L;
		else
			iratio = 100L*imagesize/packimage_size;

		if (packrelocs_size > leh.relosize)
			rratio = 100L;
		else
			rratio = 100L*leh.relosize/packrelocs_size;

		Status("All done!  Image %d.%02d:1, relocs %d.%02d:1",iratio/100,iratio%100,rratio/100,rratio%100);
	}

	return EXIT_SUCCESS;

abort:
	Error("unknown failure");
};
