#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <dos.h>
#include <malloc.h>

#include <fcntl.h>

#include <a/types.h>
#include <a/vexe.h>

//#define	MESSAGES
//#define MESSAGE_LZ

ushort	getCS(void);
ushort	getDS(void);
#pragma aux getCS = "mov ax,cs" value [ax] modify exact nomemory [ax];
#pragma aux getDS = "mov ax,ds" value [ax] modify exact nomemory [ax];

void LZDecompMem(void *dst,void *src);

uchar	dist_bits_tbl[]={
	0,0,0,0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10
};

ulong	dist_tbl[]={
	1,2,3,4,5,7,9,13,17,25,33,49,65,97,129,193,257,385,513,769,1025,1537,2049,3073
};

uchar	len_bits_tbl[]={
	0,0,0,0,0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,0
};

ulong	len_tbl[]={
	3,4,5,6,7,8,9,10,11,13,15,17,19,23,27,31,35,43,51,59,67,83
};

void LZHDecompMem(uchar *dst,uchar *src,ulong len) {
	uchar	huff_bllen_tbl[16];
	uchar	huff_len_bits[22];
	uchar	huff_ld_bits[280];
	ulong	huff_ld_first_tbl[15];
	ulong	huff_len_first_tbl[15];
	ulong	huff_ld_last_tbl[15];
	ulong	huff_len_last_tbl[15];
	ushort	huff_ld_start_tbl[15];
	ushort	huff_len_start_tbl[15];
	ushort	huff_ld_code_tbl[280];
	uchar	huff_len_code_tbl[22];
	int	i,j,k;
	ulong	accum=0;
	uchar	accum_bits=0;
	ulong	code;
	uchar	*dstend;

#ifdef MESSAGES
	puts("--- L/D bl-tree length table ---");
#endif

	for(i=0;i<16;i+=2) {
		huff_bllen_tbl[i+0] = *src & 15;
		huff_bllen_tbl[i+1] = *src>>4;
		++src;

//		printf("Length %2d: %d bits\n",i+0,huff_bllen_tbl[i+0]);
//		printf("Length %2d: %d bits\n",i+1,huff_bllen_tbl[i+1]);
	}

#ifdef MESSAGES
	puts("--- LZLength tree length table ---");
#endif
	
	for(i=0;i<22;i+=2) {
		huff_len_bits[i+0] = *src & 15;
		huff_len_bits[i+1] = *src>>4;
		++src;

//		printf("Length %2d: %d bits\n",i+0,huff_len_bits[i+0]);
//		printf("Length %2d: %d bits\n",i+1,huff_len_bits[i+1]);
	}

#ifdef MESSAGES
	puts("--- Constructing decoding system for L/D tree length table ---");
#endif

	code=0;
	k = 0;
	for(i=1;i<16;i++) {
		boolean found=FALSE;

		huff_ld_start_tbl[i-1]=k;
		huff_ld_first_tbl[i-1]=code;
		for(j=0;j<16;j++) {
			if (huff_bllen_tbl[j] == i) {
				huff_ld_code_tbl[k++] = j;
				code += 0x80000000L>>(i-1);
				found = TRUE;
			}
		}
		if (found)
			huff_ld_last_tbl[i-1]=code-1;
		else
			huff_ld_last_tbl[i-1]=0L;

//		printf("Level %2d: %08lx-%08lx, codes start at %ld\n",i,huff_ld_first_tbl[i-1],huff_ld_last_tbl[i-1],huff_ld_start_tbl[i-1]);
	}

	if (code) {
		puts("L/D-tree encoding tree is incomplete.");
		exit(5);
	}

#ifdef MESSAGES
  	puts("--- Literal/Distance tree length table ---");
#endif

	for(i=0;i<280;i++) {
		int c;

		if (accum_bits<16) {
			accum |= (((ulong)src[0]<<8) | (ulong)src[1]) << (16-accum_bits);
			accum_bits += 16;
			src += 2;
		}

		k=0;
		while(!huff_ld_last_tbl[k]) ++k;
		while(accum>huff_ld_last_tbl[k]) ++k;

//		printf("%08lx %d\n",accum,k);

		c = huff_ld_code_tbl[huff_ld_start_tbl[k]+((accum - huff_ld_first_tbl[k])>>(31-k))];

		accum_bits -= (k+1);
		accum <<= (k+1);
		huff_ld_bits[i]=c;

//		printf("Code %3d: %d bits\n",i,c);
	}

#ifdef MESSAGES
	puts("--- Constructing decoding systems for L/D and Len tables ---");
#endif

	code=0;
	k = 0;
	for(i=1;i<16;i++) {
		boolean found=FALSE;

		huff_ld_start_tbl[i-1]=k;
		huff_ld_first_tbl[i-1]=code;
		for(j=0;j<280;j++) {
			if (huff_ld_bits[j] == i) {
//				printf("\t%03x: %d bits\n",j,i);
				huff_ld_code_tbl[k++] = j;
				code += 0x80000000L>>(i-1);
				found = TRUE;
			}
		}
		if (found)
			huff_ld_last_tbl[i-1]=code-1;
		else
			huff_ld_last_tbl[i-1]=0L;
//		printf("Level %2d: %08lx-%08lx, codes start at %ld\n",i,huff_ld_first_tbl[i-1],huff_ld_last_tbl[i-1],huff_ld_start_tbl[i-1]);
	}
	if (code) {
		puts("Literal/distance tree is incomplete.");
		exit(5);
	}

	code=0;
	k = 0;
	for(i=1;i<16;i++) {
		boolean found=FALSE;

		huff_len_start_tbl[i-1]=k;
		huff_len_first_tbl[i-1]=code;
		for(j=0;j<22;j++) {
			if (huff_len_bits[j] == i) {
				huff_len_code_tbl[k++] = j;
				code += 0x80000000L>>(i-1);
				found = TRUE;
			}
		}
		if (found)
			huff_len_last_tbl[i-1]=code-1;
		else
			huff_len_last_tbl[i-1]=0L;

//		printf("Level %2d: %08lx-%08lx, codes start at %ld\n",i,huff_len_first_tbl[i-1],huff_len_last_tbl[i-1],huff_len_start_tbl[i-1]);
	}
	if (code) {
		puts("Length tree is incomplete.");
		exit(5);
	}


#ifdef MESSAGES
	puts("--- Actual data ---");
#endif

	dstend = dst+len;
	while(dst<dstend) {
		int c;

		if (accum_bits<16) {
			accum |= (((ulong)src[0]<<8) | (ulong)src[1]) << (16-accum_bits);
			accum_bits += 16;
			src += 2;
		}

//		printf("accum %08lx\n",accum);

		k=0;
		while(!huff_ld_last_tbl[k]) ++k;
		while(accum>huff_ld_last_tbl[k]) ++k;

		c = huff_ld_code_tbl[huff_ld_start_tbl[k]+((accum - huff_ld_first_tbl[k])>>(31-k))];

		accum_bits -= (k+1);
		accum <<= (k+1);

		if (c<256) {
			*dst++=c;
//			printf("bytes left: %8d   Literal %02x\n",dstend-dst,c);
		} else {
			uchar *from;
			uchar bc;

			// read distance

//			printf("Distance code %ld\n",c);

			c-=256;
			from = dst - dist_tbl[c];
			if (dist_bits_tbl[c]) {
				if (accum_bits<16) {
					accum |= (((ulong)src[0]<<8) | (ulong)src[1]) << (16-accum_bits);
					accum_bits += 16;
					src += 2;
				}
				from -= accum>>(32-dist_bits_tbl[c]);

//				printf("Xbits %ld\n",accum>>(32-dist_bits_tbl[c]));

				accum <<= dist_bits_tbl[c];
				accum_bits -= dist_bits_tbl[c];
			}

			// read length

			if (accum_bits<16) {
				accum |= (((ulong)src[0]<<8) | (ulong)src[1]) << (16-accum_bits);
				accum_bits += 16;
				src += 2;
			}

			k=0;
			while(!huff_len_last_tbl[k]) ++k;
			while(accum>huff_len_last_tbl[k]) ++k;

//			printf("Length level code: %d\n",k);

			c = huff_len_code_tbl[huff_len_start_tbl[k]+((accum - huff_len_first_tbl[k])>>(31-k))];

//			printf("Length code: %d\n",c);

			accum_bits -= (k+1);
			accum <<= (k+1);

			bc = len_tbl[c];

			if (len_bits_tbl[c]) {
				if (accum_bits<16) {
					accum |= (((ulong)src[0]<<8) | (ulong)src[1]) << (16-accum_bits);
					accum_bits += 16;
					src += 2;
				}
				bc += accum>>(32-len_bits_tbl[c]);
				accum <<= len_bits_tbl[c];
				accum_bits -= len_bits_tbl[c];
			}

			// copy

#ifdef MESSAGE_LZ
			printf("LZ: distance %ld length %d\n",dst-from,bc);
#endif

			do *dst++=*from++; while(--bc);
		}
	}
}

int main(int argc,char **argv) {
	int	i;
	struct VExeHeader vpeh;
	uchar *progmem;
	uchar *relocs;
	ulong	progsiz,relocsiz;

	if (argc<2) {
		puts(
			"VERUN V1.21, Copyright (C) 1997 by Avery Lee.  All Rights Reserved.\n\n"
			"\tVERUN <VE executable>\n"
			"\n"
			"VERUN is freely distributable under the terms of the GNU General\n"
			"Public License Version 2; see the VE archive for more details.\n"
		);
		return EXIT_FAILURE;
	}

	// cut the name off the cmd tail

	{
		char __far *s,__far *t,__far *u;

		s = t = u = MK_FP(_psp,0x81);

		u = s + *(uchar far *)(_psp:>0x80);

		while(s<u)
			if (isspace(*s)) ++s; else break;

		while(s<u)
			if (!isspace(*s)) ++s; else break;

		while(s<u)
			*t++ = *s++;

		*(uchar far *)(_psp:>0x80) = FP_OFF(t)-0x81;
	}

	if (-1L == (i = open(argv[1],O_RDONLY+O_BINARY))) goto readerr;
	
	if (sizeof vpeh != read(i,&vpeh,sizeof vpeh)) goto readerr;

	if (vpeh.ID != VEXE_ID) {
		puts("This is not an unbound VE executable.");
		return EXIT_FAILURE;
	}

	progsiz = vpeh.uc_image_length+max(18,vpeh.bss_length);
	relocsiz = vpeh.uc_relocs_length+18;

	if (!(progmem = malloc(progsiz))) goto memerr;
	if (!(relocs = malloc(relocsiz))) goto memerr;

	if (0>lseek(i,vpeh.image_offset,SEEK_SET)) goto readerr;

	switch(vpeh.image_comp) {
	case VEXE_COMP_LZH:
		if (vpeh.image_length != read(i,progmem + progsiz - vpeh.image_length,vpeh.image_length)) goto readerr;
		LZHDecompMem(progmem,progmem + progsiz - vpeh.image_length,vpeh.uc_image_length);
		break;
	case VEXE_COMP_LZSS:
		if (vpeh.image_length != read(i,progmem + progsiz - vpeh.image_length,vpeh.image_length)) goto readerr;
		LZDecompMem(progmem,progmem + progsiz - vpeh.image_length);
		memmove(progmem,progmem+18,vpeh.uc_image_length);
		break;
	case VEXE_COMP_NONE:
		if (vpeh.image_length != read(i,progmem,vpeh.image_length)) goto readerr;
		break;
	default:
		puts("Unknown relocs compression type.");
		return EXIT_FAILURE;
	}

	if (0>lseek(i,vpeh.relocs_offset,SEEK_SET)) goto readerr;

	switch(vpeh.relocs_comp) {
	case VEXE_COMP_LZH:
		if (vpeh.relocs_length != read(i,relocs + relocsiz - vpeh.relocs_length,vpeh.relocs_length)) goto readerr;
		LZHDecompMem(relocs,relocs+relocsiz-vpeh.relocs_length,vpeh.uc_relocs_length);
		break;
	case VEXE_COMP_LZSS:
		if (vpeh.relocs_length != read(i,relocs + relocsiz - vpeh.relocs_length,vpeh.relocs_length)) goto readerr;
		LZDecompMem(relocs,relocs + relocsiz - vpeh.relocs_length);
		memmove(relocs,relocs+18,vpeh.uc_relocs_length);
		break;
	case VEXE_COMP_NONE:
		if (vpeh.relocs_length != read(i,relocs,vpeh.relocs_length)) goto readerr;
		break;
	default:
		puts("Unknown relocs compression type.");
		return EXIT_FAILURE;
	}

	close(i);

	// clear BSS area

	memset(progmem + vpeh.uc_image_length,0,vpeh.bss_length);

#ifdef MESSAGES
	printf("Uncompressed image length: %08lx bytes (%08lx BSS)\n",vpeh.uc_image_length,vpeh.bss_length);
#endif

	// apply relocs to program

//printf("entry pt: %08lx\n",progmem + vpeh.entry_point);

	{
		uchar	*rp = relocs;
		uchar	*pp;
		uchar	type;
		ulong	num;
		ulong	diff;
#ifdef MESSAGES
		ulong	numrelocs = 0;
#endif

		while(type = *rp++) {
			if (type>0x03) {
//				printf("Invalid reloc type: %02x",type);
				puts("Invalid reloc type!");
				return EXIT_FAILURE;
			}
			num = *(ushort *)rp;	rp+=2;
			if (!num) num = 0x10000L;
			pp = progmem + *(ulong *)rp;	rp+=4;

//			printf("Applying reloc type %02x to %ld relocs\n",type,num);

			do {
				if (type & 1) {
//printf("%08lx %08lx %08lx\n",pp-progmem,*(ulong *)pp,progmem+*(ulong *)pp);
					*(ulong *)pp += (ulong)progmem;
					pp+=4;
				}
				if (type & 2) {
					if (*(ushort *)pp)
						*(ushort *)pp = getCS();
					else
						*(ushort *)pp = getDS();

					pp+=2;
				}
#ifdef MESSAGES
			++numrelocs;
#endif
				if (!--num) break;
				diff=0;
				do {
					diff<<=7;
					diff += (ulong)(*rp & 0x7f);
				} while(*rp++ & 0x80);
				pp += diff;
			} while(num);
		}
#ifdef MESSAGES
		printf("Applied %ld relocs.\n",numrelocs);
#endif
	}

	free(relocs);

//	puts("Reloc finished!");

//	((void (*)(void))(progmem + vpeh.entry_point))();

//	printf("Entry point: %08lx  Stack point: %08lx\n",vpeh.entry_point,vpeh.stack_point);

#ifdef MESSAGES
	puts("Ready to execute.");
	getch();
#endif

	run_ve_asm(progmem+vpeh.entry_point,progmem+vpeh.stack_point);

	return EXIT_SUCCESS;
readerr:
	puts("Error reading VE exe.");
	return EXIT_FAILURE;
memerr:
	puts("Not enough memory.");
	return EXIT_FAILURE;
}


